/*
* Copyright (c) 2012 Paweł Panasewicz
*
*    Licensed under the Apache License, Version 2.0 (the "License");
*    you may not use this file except in compliance with the License.
*    You may obtain a copy of the License at
*
*        http://www.apache.org/licenses/LICENSE-2.0
*
*    Unless required by applicable law or agreed to in writing, software
*    distributed under the License is distributed on an "AS IS" BASIS,
*    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*    See the License for the specific language governing permissions and
*    limitations under the License.
*/

package pl.panasoft.pimps

import language.implicitConversions
import com.weiglewilczek.slf4s.{Logger, Logging}
import util.Random
import scalaz._
import Scalaz._
import org.slf4j.LoggerFactory


object LogAround {

  def makeStamp = new Random().alphanumeric.take(10).mkString

  def traceAround[E, A](a: => A, subject: => E, logger: Logger): A = logAround(a, subject, "trace", logger)

  def debugAround[E, A](a: => A, subject: => E, logger: Logger): A = logAround(a, subject, "debug", logger)

  def infoAround[E, A](a: => A, subject: => E, logger: Logger): A = logAround(a, subject, "info", logger)

  def warnAround[E, A](a: => A, subject: => E, logger: Logger): A = logAround(a, subject, "warn", logger)

  def errorAround[E, A](a: => A, subject: => E, logger: Logger): A = logAround(a, subject, "error", logger)

  private def logAround[E, A](a: => A, subject: => E, methodName: String, logger: Logger): A = {

    def log(m: => String) = methodName match {
      case "trace" => logger.trace(m)
      case "debug" => logger.debug(m)
      case "info" => logger.info(m)
      case "warn" => logger.warn(m)
      case "error" => logger.error(m)
    }

    val stamp = makeStamp

    log(s"$subject STARTED <logaround:$stamp>")
    val c = a
    c match {
      case Success(s) => log(s"$subject FINISHED-Successful </logaround:$stamp>")
      case Failure(f) => log(s"$subject FINISHED-Failed ($f) </logaround:$stamp>")
      case _ => log(s"$subject FINISHED </logaround:$stamp>")
    }
    a
  }
}

trait LogAroundOps[A] extends syntax.Ops[A] {

  implicit def logger: Logger

  def traceAround[E](subject: => E): A = LogAround.traceAround(self, subject, logger)

  def debugAround[E](subject: => E): A = LogAround.debugAround(self, subject, logger)

  def infoAround[E](subject: => E): A = LogAround.infoAround(self, subject, logger)

  def warnAround[E](subject: => E): A = LogAround.warnAround(self, subject, logger)

  def errorAround[E](subject: => E): A = LogAround.errorAround(self, subject, logger)
}

trait ToLogAroundOps {

  implicit val logger = Logger("DefaultLogAround")

  implicit def ToLogAroundOps[A](a: => A)(implicit F: Logger) = new LogAroundOps[A] {
    def self = a
    def logger = F
  }
}

trait ImplicitLogger {
  implicit val logger = Logger(this.getClass)
}